# coding=utf8
"""REST

Extends Service to allow for launching as an http accesible rest
service
"""
from __future__ import annotations

__author__		= "Chris Nasr"
__copyright__	= "Ouroboros Coding Inc."
__email__		= "chris@ouroboroscoding.com"
__created__		= "2023-03-17"

# Limit exports
__all__ = [ 'bottle', 'REST' ]

# Ouroboros imports
from jobject import jobject
import jsonb
import memory
from tools import clone

# Python imports
from collections.abc import Callable
from datetime import datetime
import re
import sys
import traceback
from typing import List, Literal, TYPE_CHECKING

# Pip imports
import bottle

# Local imports
from body.errors import \
	REST_AUTHORIZATION, REST_CONTENT_TYPE, REST_LIST_INVALID_URI, \
	REST_LIST_TO_LONG, REST_REQUEST_DATA, SERVICE_CRASHED, SERVICE_NO_DATA, \
	SERVICE_NO_SESSION
from body.response import Error, Response, ResponseException
if TYPE_CHECKING:
	from body.service import Service

class _Route(object):
	"""Route

	A private callable class used to store rest routes accessed by bottle
	"""

	__content_type = re.compile(r'^application\/json; charset=utf-?8$')
	"""Valid Content-Type"""

	__cors = None
	"""CORs regular expression"""

	__key_to_errors = {
		'data': SERVICE_NO_DATA,
		'session': SERVICE_NO_SESSION
	}
	"""Maps key error variables to their response error code"""

	__on_error = None
	"""On Error
	Function called when a service request raises an exception
	"""

	__services = []
	"""Services
	Minimizes space usage on service names repeating
	"""

	__uris: dict = {}
	"""URIs
	used to keep track of the uris to callbacks for the purposes of __list calls
	"""

	__verbose = True
	"""The verbose mode, set to True to view requests/responses"""

	@classmethod
	def cors(cls, cors):
		"""CORs

		Sets the regular expression used to validate domains making requests

		Arguments:
			pattern (Re.Pattern): A compiled regular expression

		Returns:
			None
		"""
		cls.__cors = cors

	@classmethod
	def on_error(cls, callback: callable):
		"""On Error

		Sets the global error handling function

		Arguments:
			callback (callable): The function to call with the error details

		Returns:
			None
		"""
		cls.__on_error = staticmethod(callback)

	@classmethod
	def verbose(cls, b: bool):
		"""Verbose

		Set the verbose mode associated with all routes

		Arguments:
			b (bool): True/False

		Returns:
			None
		"""
		cls.__verbose = b

	def __init__(self,
		service: str,
		callback: callable | Literal[True],
		uri: str | None = None
	):
		"""Constructor

		Initialises an instance of the route

		Arguments:
			service (str): The name of the service associated with this route
			callback (callable): The function to pass details to when this \
				route is triggered. Send True to make a __list route
			uri (str):

		Returns:
			None
		"""

		# Set the name for the route
		self.__name__ = callback == True and 'True' or str(callback)

		# Store the callback
		self.__callback = callback

		# Get the index of the service
		try:
			self._service = self.__services.index(service)
		except ValueError:
			self.__services.append(service)
			self._service = len(self.__services) - 1

		# If we got a uri, associate the callback with it globally, this way
		#	we can reference it when getting __list requests
		if uri is not None:
			self.__uris[uri[1:]] = callback

	def __call__(self):
		"""Call (__call__)

		Python magic method that allows the instance to be called

		Returns:
			str
		"""

		# If CORS is enabled and the origin matches
		if self.__cors and \
			'origin' in bottle.request.headers and \
			self.__cors.match(bottle.request.headers['origin']):

			# Add the current origin as acceptable
			bottle.response.headers['Access-Control-Allow-Origin'] = \
				bottle.request.headers['origin']
			bottle.response.headers['Vary'] = 'Origin'

		# If the request is OPTIONS
		if bottle.request.method == 'OPTIONS':

			# If we're in verbose mode, let people know we are handling OPTIONS
			if self.__verbose:
				print('%s: %s OPTIONS %s' % (
					str(datetime.now()),
					self.__services[self._service],
					bottle.request.path
				))

			# Set the default headers expected for OPTIONS
			bottle.response.headers['Access-Control-Allow-Methods'] = \
				'DELETE, GET, POST, PUT, OPTIONS'
			bottle.response.headers['Access-Control-Max-Age'] = 1728000
			bottle.response.headers['Access-Control-Allow-Headers'] = \
				'Authorization,DNT,X-CustomHeader,Keep-Alive,User-Agent,' \
				'X-Requested-With,If-Modified-Since,Cache-Control,Content-Type'
			bottle.response.headers['Content-Type'] = 'text/plain charset=UTF-8'
			bottle.response.headers['Content-Length'] = 0
			bottle.request.status = 204
			return ''

		# Set the return to JSON
		bottle.response.headers['Content-Type'] = \
			'application/json; charset=utf-8'

		# Initialise the request details with the bottle request and response
		#	objects, allowing requests direct access to low level data
		oReq = jobject({ })

		# If we got a Read request and the data is in the GET
		if bottle.request.method == 'GET' and 'd' in bottle.request.query:

			# Convert the GET and store the data
			try:
				oReq.data = jsonb.decode(bottle.request.query['d'])
			except Exception as e:
				return Error(
					REST_REQUEST_DATA,
					'%s\n%s' % ( bottle.request.query['d'], str(e) )
				).to_json()

		# Else we most likely got the data in the body
		else:

			# Make sure the request send JSON
			try:
				if not self.__content_type.match(
					bottle.request.headers['Content-Type'].lower()
				):
					return str(Error(REST_CONTENT_TYPE))
			except KeyError:
				return Error(REST_CONTENT_TYPE).to_json()

			# Store the data, if it's too big we need to read it rather than
			#	use getvalue
			try:
				sData = bottle.request.body.getvalue()
			except AttributeError as e:
				sData = bottle.request.body.read()

			# Make sure we have a string, not a set of bytes
			try:
				sData = sData.decode()
			except (UnicodeDecodeError, AttributeError):
				pass

			# Convert the data and store it
			try:
				if sData: oReq.data = jsonb.decode(sData)
			except Exception as e:
				return Error(
					REST_REQUEST_DATA,
					'%s\n%s' % ( sData, str(e) )
				).to_json()

		# If the request sent a authorization token
		if 'Authorization' in bottle.request.headers:

			# Get the session from the Authorization token
			oReq.session = memory.load(bottle.request.headers['Authorization'])

			# If the session is not found
			if not oReq.session:
				bottle.response.status = 401
				return Error(
					REST_AUTHORIZATION, 'Unauthorized'
				).to_json()

			# Else, extend the session's ttl
			else:
				oReq.session.extend()

		# Step through all headers
		for k in bottle.request.headers:
			if k[0:7] == 'X-Body-':
				try:
					oReq.meta[k[7:]] = bottle.request.headers[k]
				except AttributeError:
					oReq.meta = { k[7:]: bottle.request.headers[k] }

		# In case the service crashes
		try:

			# If we're in verbose mode
			if self.__verbose:
				print('%s REQUEST %s %s %s %s' % (
					str(datetime.now()),
					self.__services[self._service],
					bottle.request.method,
					bottle.request.path,
					(oReq and jsonb.encode(oReq, 2) or 'None')
				))

			# If this is a list request
			if self.__callback is True:

				# If the data isn't passed or isn't an array
				if 'data' not in oReq or not isinstance(oReq.data, list):
					oResponse = Error(
						REST_REQUEST_DATA,
						'data must be an array'
					)

				# Else, if it's beyond the max
				elif len(oReq.data) > 10:
					oResponse = Error(
						REST_LIST_TO_LONG,
						'Can not request more than 10 urls via __list'
					)

				# Else, we have an array of nouns to call and return
				else:

					# Init the response
					lResponse = []

					# Go through each element in the list
					for m in oReq.data:

						# If we got a string
						if isinstance(m, str):
							m = [ m ]

						# Else, if we didn't get a list
						elif not isinstance(m, list):
							oResponse = Error(
								REST_REQUEST_DATA, [ m,
									'data must be an array or URI and data, ' \
									'or single string for the URI' ]
							)
							break

						# If the URI doesn't exist
						if m[0] not in self.__uris:
							oResponse = Error(
								REST_LIST_INVALID_URI,
								m[0]
							)
							break

						# Clone the __list request data so that the session
						#	or any other data is shared
						oRequest = clone(oReq)

						# If unique data was passed for the child request
						if len(m) == 2:

							# If we didn't get a dict
							if not isinstance(m[1], dict):
								oResponse = Error(
									REST_LIST_INVALID_URI,
									[ m[1], 'data must be an object' ]
								)
								break

							# Set the data for this request
							oRequest.data = m[1]

						# No unique data was sent, so we need to delete the
						#	original data from the request
						else:
							try: del oRequest.data
							except AttributeError: pass

						# Call the request and append the data to the
						#	response
						try:
							lResponse.append([
								m[0],
								self.__uris[m[0]](oRequest).to_dict()
							])

						# If we got a KeyError
						except (AttributeError, KeyError) as e:
							if e.args[0] in self.__key_to_errors:
								oResponse = Error(
									self.__key_to_errors[e.args[0]]
								)
								break
							else:
								raise

						# If we got a response exception
						except ResponseException as e:

							# Set the response using the exceptions first argument
							oResponse = e.args[0]
							break

					# Else, we got through the list
					else:

						# Set the response using the list of individual
						#	responses
						oResponse = Response(lResponse)

			# Else, we are making a single URI request
			else:

				# Call the appropriate API method based on the HTTP/request
				#	method
				try:
					oResponse = self.__callback(oReq)

				# If we got a KeyError
				except (AttributeError, KeyError) as e:
					if e.args[0] in self.__key_to_errors:
						oResponse = Error(
							self.__key_to_errors[e.args[0]]
						)
					else:
						raise

				# If we got a response exception
				except ResponseException as e:

					# Set the response using the exceptions first argument
					oResponse = e.args[0]

		# If we get absolutely any exception
		except Exception as e:

			# Get the traceback info
			sError = traceback.format_exc()

			# Print the traceback to stderr
			print(sError, file = sys.stderr)

			# If we have an error handler
			if self.__on_error:

				# Gather all the details, including optional ones
				oDetails = {
					'service': self.__services[self._service],
					'method': bottle.request.method,
					'path': bottle.request.path,
					'environment': bottle.request.environ,
					'traceback': sError
				}
				for s in [ 'data', 'session' ]:
					if s in oReq:
						oDetails[s] = oReq[s]

				# Pass the details to the error handler
				self.__on_error(oDetails)

			# Set a response of service/request crashed
			oResponse = Error(
				SERVICE_CRASHED,
				'%s:%s' % ( self.__services[self._service], bottle.request.path )
			)

		# If the response contains an error
		if oResponse.error:

			# If it's an authorization error
			if oResponse.error['code'] == REST_AUTHORIZATION:

				# Set the http status to 401 Unauthorized
				bottle.response.status = 401

				# If the message is missing
				if oResponse.error['msg'] == '':
					oResponse.error['msg'] = 'Unauthorized'

			# Add the service and path to the call
			l = [
				self.__services[self._service],
				bottle.request.method,
				bottle.request.path
			]
			try:
				oResponse.error['service'].append(l)
			except KeyError:
				oResponse.error['service'] = [ l ]

		# If we're in verbose mode
		if self.__verbose:
			print('%s RETURNING %s %s %s %s' % (
				str(datetime.now()),
				self.__services[self._service],
				bottle.request.method,
				bottle.request.path,
				jsonb.encode(oResponse.to_dict(), 2)
			)
		)

		# Return the Response as a string
		return oResponse.to_json()

class REST(bottle.Bottle):
	"""REST

	Used to access a service instance via HTTP/REST

	Extends:
		Bottle
	"""

	__action_to_method = {
		'create': 'POST',
		'delete': 'DELETE',
		'read':   'GET',
		'update': 'PUT'
	}
	"""Maps HTTP methods to service actions"""

	def __init__(self,
		instances: List[Service],
		cors: List[str] | None = None,
		lists: str | Literal[True] = True,
		on_errors: Callable | None = None,
		verbose: bool = False
	):
		"""Constructor

		Creates a new REST instance

		Arguments:
			instances (dict): The service names to instances to make accessible
				via REST
			cors (str[]): A list of allowed domains for CORS policy
			lists (str | True): True to add `__list` to each service, else a str
				to use instead of `__list`
			on_errors (callable): Optional, a function to call when a service
				request throws an exception
			verbose (bool): Optional, set to True to print out each request and
				and response

		Raises:
			ValueError

		Returns:
			RestService
		"""

		# Call the parent constructor first so the object is setup
		super(REST, self).__init__()

		# If the instances are not a list
		if not isinstance(instances, list):
			raise TypeError(
				'instances', 'must be a list', sys._getframe().f_code.co_name
			)

		# If cors, compile it
		if cors:

			# If it's not a list
			if not isinstance(cors, list):
				raise RuntimeError('REST.cors must be a list')

			# If we only have one
			if len(cors) == 1:
				cors = cors[0].replace('.', '\\.')
			else:
				cors = '(?:%s)' % '|'.join([
					s.replace('.', '\\.')
					for s in cors
				])
			cors = re.compile('https?://(.*\\.)?%s' % cors)

			# Set it
			_Route.cors(cors)

		# If we have an error handler
		if on_errors:
			_Route.on_error(on_errors)

		# Set the verbose mode
		_Route.verbose(verbose)

		# Step through each service
		bOne = len(instances) == 1
		for oInstance in instances:

			# Go through all the request methods found in the service
			for dRequest in oInstance.requests:

				# Get the method
				sMethod = self.__action_to_method[dRequest['action']]

				# Generate the URI
				sUri = '/%s' % dRequest['name'].replace('_', '/')

				# Register it with bottle
				self.route(
					bOne and sUri or ('/%s%s' % ( oInstance.name, sUri )),
					[ sMethod, 'OPTIONS' ],
					_Route(
						oInstance.name,
						dRequest['func'],
						uri = (list and sMethod == 'GET') and sUri or None
					)
				)

			# If we have additional static routes
			if oInstance.additional_routes is not None:
				for l in oInstance.additional_routes:
					self.route(
						bOne and l[0] or f'/{oInstance.name}{l[0]}',
						isinstance(l[1], list) \
							and l[1][:] + [ 'OPTIONS' ] \
							or [ l[1], 'OPTIONS' ],
						l[2]
					)

			# If we have a request for a list of requests
			if lists:

				# If it's True
				sList = '/%s' % (lists is True and '__list' or str(lists))

				# Add the list read route
				self.route(
					bOne and sList or ('/%s%s' % ( oInstance.name, sList)),
					[ 'GET', 'OPTIONS' ],
					_Route(oInstance.name, True)
				)

	# run method
	def run(self, server = 'gunicorn', host = '127.0.0.1', port = 8080,
			reloader = False, interval = 1, quiet = False, plugins = None,
			debug = None, maxfile = 20971520, **kargs):
		"""Run

		Overrides Bottle's run to default gunicorn and other fields

		Arguments:
			server (str): Server adapter to use
			host (str): Server address to bind to
			port (int): Server port to bind to
			reloader (bool): Start auto-reloading server?
			interval (int): Auto-reloader interval in seconds
			quiet (bool): Suppress output to stdout and stderr?
			plugins (list): List of plugins to the server
			debug (bool): Debug mode
			maxfile (int): Maximum size of requests

		Returns:
			None
		"""

		# Set the max file size
		bottle.BaseRequest.MEMFILE_MAX = maxfile

		# Call bottle run
		bottle.run(
			app = self, server = server, host = host, port = port,
			reloader = reloader, interval = interval, quiet = quiet,
			plugins = plugins, debug = debug, **kargs
		)


# coding=utf8
""" Response

Holds the class that holds Response messages sent back from Service Rerquests
"""

__author__		= "Chris Nasr"
__copyright__	= "Ouroboros Coding Inc."
__email__		= "chris@ouroboroscoding.com"
__created__		= "2023-03-15"

# Ouroboros imports
import jsonb
import undefined

class Response(object):
	"""Response

	Represents a standard result from any/all requests
	"""

	def __init__(self,
		data: any = undefined,
		error: any = undefined,
		warning: any = undefined
	):
		"""Constructor

		Initialises a new Response instance

		Arguments:
			data (mixed): If a request returns data this should be set
			error (mixed): If a request has an error, this can be filled with
				a code and message string
			warning (mixed): If a request returns a warning this should be set

		Raises:
			ValueError

		Returns:
			Response
		"""

		# If there's no data
		if data is not undefined:
			self.data = data
		else:
			self.data = None

		# If there's an error, figure out what type
		if error is not undefined:

			# If we got an int, it's a code with no message string
			if isinstance(error, int):
				self.error = {'code': error, 'msg': ''}

			# If we got a string, it's a message with no code
			elif isinstance(error, str):
				self.error = {'code': 0, 'msg': error}

			# If it's a tuple, 0 is a code, 1 is a message
			elif isinstance(error, tuple):
				self.error = {'code': error[0], 'msg': error[1]}

			# If we got a dictionary, assume it's already right
			elif isinstance(error, dict):
				self.error = error

			# If we got an exception
			elif isinstance(error, Exception):

				# If we got another Response in the Exception, store the error
				#	from it
				if isinstance(error.args[0], Response):
					self.error = error.args[0].error

				# Else, try to pull out the code and message
				else:
					self.error = {'code': error.args[0], 'msg': ''}
					if len(error.args) > 1: self.error['msg'] = error.args[1]

			# Else, we got something invalid
			else:
				raise ValueError('error', 'invalid value: %s' % str(error))

		# Else, set it to False
		else:
			self.error = False

		# If there's a warning, store it as is
		if warning is not undefined:
			self.warning = warning

	def __bool__(self):
		"""bool

		Python magic method to return a bool from the instance. In the case of
		response, True is returned if there's data, else False

		Returns:
			boolean
		"""
		try: return self.data != None
		except AttributeError: return False

	def __repr__(self):
		"""repr

		Python magic method to return a string from the instance that can be
		turned back into the instance

		Returns:
			str
		"""
		return str(self.to_dict())

	def __str__(self):
		"""str

		Python magic method to return a string from the instance

		Returns:
			str
		"""
		return str(self.to_dict())

	@classmethod
	def from_dict(cls, val):
		"""From Dict

		Converts a dict back into an Response

		Arguments:
			val (dict): A valid dict

		Returns:
			Response
		"""

		# Create a new instance
		o = cls()

		# If there's data (more likely to be there, so try/except)
		try:
			o.data = val['data']
		except KeyError:
			o.data = None

		# If there's an error (less likely, use if/in)
		if 'error' in val:
			o.error = val['error']
		else:
			o.error = False

		# If there's a warning (less likely, use if/in)
		if 'warning' in val:
			o.warning = val['warning']

		# Return the instance
		return o

	@classmethod
	def from_json(cls, val):
		"""From JSON

		Tries to convert a string made from str() back into an Response

		Arguments:
			val (str): A valid JSON string

		Returns:
			Response
		"""

		# Try to convert the string to a dict
		try: d = jsonb.decode(val)
		except ValueError as e: raise ValueError('val', str(e))
		except TypeError as e: raise ValueError('val', str(e))

		# Return the fromDict result
		return cls.from_dict(d)

	def to_dict(self):
		"""To Dict

		Converts the Response into a dict

		Returns:
			dict
		"""

		# Init the return
		dRet = {}

		# Set the data key if it's set in the instance
		if self.data is not None:
			dRet['data'] = self.data

		# If the error is not set to False
		if self.error is not False:
			dRet['error'] = self.error

		# Look for a warning attribute (less likely, use if/hasattr)
		if hasattr(self, 'warning'):
			dRet['warning'] = self.warning

		# Return the dict
		return dRet

	def to_json(self):
		"""To JSON

		Returns a JSON representation of the object

		Returns:
			str
		"""
		return jsonb.encode(self.to_dict())

	def warning_exists(self):
		"""Warning Exists

		Returns True if there is a warning in the Response

		Returns:
			bool
		"""
		return hasattr(self, 'warning')

class Error(Response):
	"""Error

	Shorthand form of Response(error=)
	"""

	def __init__(self, code, msg=None):
		"""Constructor

		Initialises a new Response instance

		Arguments:
			code (uint): The error code
			msg (mixed): Optional message for more info on the error

		Returns:
			Error
		"""

		# Set the data to None
		self.data = None

		# Set the error code
		self.error = {
			'code': code,
			'msg': msg
		}

class ResponseException(Exception):
	"""Response Exception

	Python won't let you raise anything that doesn't extend BaseException
	"""

	def __init__(self,
		data: any = undefined,
		error: any = undefined,
		warning: any = undefined
	):
		"""Constructor

		Creates a new instance of the Exception

		Arguments:
			data (mixed): If a request returns data this should be set
			error (mixed): If a request has an error, this can be filled with
				a code and message string
			warning (mixed): If a request returns a warning this should be set

		Returns:
			ResponseException
		"""

		# If we got a Response object
		if isinstance(data, Response):
			super().__init__(data)

		# Else, construct the Response and pass it to the parent
		else:
			super().__init__(Response(data, error, warning))


# coding=utf8
"""External

Methods to connecting to external services via http
"""
from __future__ import annotations

__author__		= "Chris Nasr"
__copyright__	= "Ouroboros Coding Inc."
__email__		= "chris@ouroboroscoding.com"
__created__		= "2025-04-06"

__all__ = [
	'create', 'delete', 'read', 'register_service', 'request', 'service_info',
	'update'
]

# Ouroboros imports
from config import config
from jobject import jobject
import jsonb
import undefined

# Python imports
from copy import copy
from time import sleep

# Pip imports
import requests
from typing import TYPE_CHECKING, MutableMapping

# Local imports
from body import errors
from body.response import Error, Response, ResponseException
if TYPE_CHECKING:
	from body.service import Service

__services = None
"""Registered Services"""

__action_to_request = {
	'create': [ requests.post, 'POST' ],
	'delete': [ requests.delete, 'DELETE' ],
	'read': [ requests.get, 'GET' ],
	'update': [ requests.put, 'PUT' ]
}
"""Map actions to request methods"""

def create(
	service: str,
	path: str,
	req: dict = {}
):
	"""Create

	Make a POST request

	Arguments:
		service (str): The service to call
		path (str): The path on the service
		req (dict): The request details, which can include 'data' and 'session'

	Returns:
		Response
	"""
	return request(service, 'create', path, req)

def delete(
	service: str,
	path: str,
	req: dict = {}
):
	"""Delete

	Make a DELETE request

	Arguments:
		service (str): The service to call
		path (str): The path on the service
		req (dict): The request details, which can include 'data' and 'session'

	Returns:
		Response
	"""
	return request(service, 'delete', path, req)

def read(
	service: str,
	path: str,
	req: dict = {}
):
	"""Read

	Make a GET request

	Arguments:
		service (str): The service to call
		path (str): The path on the service
		req (dict): The request details, which can include 'data' and 'session'

	Returns:
		Response
	"""
	return request(service, 'read', path, req)

def register_service(name: str, instance: Service):
	"""Register Service

	Marks a service as internal so that any create, read, update, delete request
	will call the local instance instead of attempting an http request

	Arguments:
		name (str): The name of the service and how others will contact it
		instance (Service): The created instance of the servce
	"""

	global __services

	# If we haven't already, generate the list of services
	if __services is None:
		_generate_services()

	# Add the instance and init the paths
	__services[name]['instance'] = instance
	__services[name]['paths'] = {}

	# Go through all the request methods in the service
	for dMethod in instance.requests:

		# Generate the URL
		sUri = dMethod['name'].replace('_', '/')

		# Store the url and action
		try:
			__services[name]['paths'][sUri][dMethod['action']] = dMethod['func']
		except KeyError:
			__services[name]['paths'][sUri] = {
				dMethod['action']: dMethod['func']
			}

	# Return the service
	return __services[name]

def _generate_services():
	"""Generate Services

	Fetches the service information from the config and generates the URLs for
	all available services
	"""

	# Pull in the global services
	global __services

	# Fetch the REST config
	dRest = config.body.rest({
		'allowed': None,
		'default': {
			'domain': 'localhost',
			'host': '0.0.0.0',
			'port': 9000,
			'protocol': 'http',
			'workers': 1
		}
	})

	# If we didn't get a list of services
	if 'services' not in dRest:
		raise ValueError('config.body.rest.services', 'missing')

	# Init the defaults if none are found
	if 'defaults' not in dRest:
		dRest['defaults'] = {}

	# Port values are not modified by default
	iPortMod = 0

	# If there is a port modifier
	if 'port' in dRest['default']:

		# Make sure it's an integer
		try:
			iPortMod = int(dRest['default']['port'])
			del dRest['default']['port']
		except ValueError:
			raise ValueError('config.body.rest.default.port', 'must be an int')

	# Reset the dict of services
	__services = {}

	# Loop through the list of services in the rest config
	for s in dRest['services']:

		# If the service doesn't point to a dict
		if not isinstance(dRest['services'][s], dict):
			raise ValueError(
				'config.body.rest.services.%s' % s, 'must be a dict'
			)

		# Start with the default values
		dParts = dRest['default'].copy()

		# Then add the service values
		dParts.update(dRest['services'][s])

		# If we have no port
		if 'port' not in dParts:

			# But we have a modifier, assume we add to 80
			if iPortMod: dParts['port'] = 80 + iPortMod

		# Else add the modifier to the port passed
		else:
			dParts['port'] += iPortMod

		# Set defaults for any missing parts
		if not dParts['protocol']: dParts['protocol'] = 'http'
		if not dParts['domain']: dParts['domain'] = 'localhost'
		if 'path' not in dParts: dParts['path'] = ''
		else: dParts['path'] = '%s/' % str(dParts['path'])

		# Store the parts for the service
		__services[s] = dParts.copy()

		# Generate a URL from the parts and store it
		__services[s]['url'] = '%s://%s%s/%s' % (
			dParts['protocol'],
			dParts['domain'],
			'port' in dParts and ":%d" % dParts['port'] or '',
			dParts['path']
		)

		# If we still have no port, default to 80
		if 'port' not in __services[s]:
			__services[s]['port'] = 80

def request(
	service: str,
	action: str,
	path: str,
	req: MutableMapping = {}
):
	"""Request

	Method to convert REST requests into HTTP requests

	Arguments:
		service (str): The service we are requesting data from
		action (str): The action to take on the service
		path (str): The path of the request
		req (dict): The request details: 'data', 'session', and 'enviroment'

	Raises:
		KeyError: if the service or action don't exist

	Return:
		Response
	"""

	global __services

	# If we haven't already, generate the list of services
	if __services is None:
		_generate_services()

	# Init the data and headers
	sData = ''
	dHeaders = {}

	# Add the default content length and type
	dHeaders['Content-Length'] = '0'
	dHeaders['Content-Type'] = 'application/json; charset=utf-8'

	# If the data was passed
	if 'data' in req and req['data']:

		# Convert the data to JSON and store the length
		sData = jsonb.encode(req['data'])
		dHeaders['Content-Length'] = str(len(sData))

	# If we have a session, add the ID to the headers
	if 'session' in req and req['session']:
		dHeaders['Authorization'] = req['session'].key()

	# If we got a service instance
	if 'instance' in __services[service]:

		# Try to find the method
		try:
			f = __services[service]['paths'][path][action]

		# If we got a KeyError
		except KeyError as e:
			if e.args[0] == path:
				return Error(errors.SERVICE_NO_SUCH_NOUN)
			elif e.args[0] == action:
				return Error(errors.SERVICE_ACTION)
			else:
				raise e

		# Try to call the method
		try:
			return f(jobject(req))

		# If we got a KeyError
		except (AttributeError, KeyError) as e:
			if e.args[0] == 'data':
				return Error(errors.SERVICE_NO_DATA)
			elif e.args[0] == 'session':
				return Error(errors.SERVICE_NO_SESSION)
			else:
				raise e

		# If we got a response exception, return the Response
		except ResponseException as e:
			return e.args[0]

	# Else, this is an external service
	else:

		# If we received any meta vars
		if 'meta' in req and req['meta']:
			for k,v in req['meta'].items():
				dHeaders['X-Body-%s' % k] = v

		# Loop requests so we don't fail just because of a network hiccup
		iAttempts = 0
		while True:

			# Increase the attempts
			iAttempts += 1

			# Make the request using the services URL and the current path, then
			#	store the response
			try:
				oRes = __action_to_request[action][0](
					__services[service]['url'] + path,
					data = sData,
					headers = dHeaders
				)

				# If the request wasn't successful
				if oRes.status_code != 200:

					# If we got a 401
					if oRes.status_code == 401:
						return Response.from_json(oRes.content)
					else:
						return Error(
							errors.SERVICE_STATUS,
							'%d: %s' % (oRes.status_code, oRes.content)
						)

				# If we got the wrong content type
				if oRes.headers['Content-Type'].lower() != \
					'application/json; charset=utf-8':
					return Error(
						errors.SERVICE_CONTENT_TYPE,
						'%s' % oRes.headers['content-type']
					)

				# Turn the content into a Response and return it
				return Response.from_json(oRes.text)

			# If we couldn't connect to the service
			except requests.ConnectionError as e:

				# If we haven't exhausted attempts
				if iAttempts < 3:

					# Wait for a second
					sleep(1)

					# Loop back around
					continue

				# We've tried enough, return an error
				return Error(errors.SERVICE_UNREACHABLE, str(e))

def service_info(name: str) -> dict:
	"""Service Info

	Returns the info related to a specific service

	Arguments:
		name (str): The name of the service

	Raises:
		KeyError if the name doesn't match a service

	Returns:
		dict
	"""

	global __services

	# If we haven't already, generate the list of services
	if __services is None:
		_generate_services()

	# Return a shallow copy of the info for the service
	return copy(__services[name])

def update(
	service: str,
	path: str,
	req: dict = {}
):
	"""Update

	Make a PUT request

	Arguments:
		service (str): The service to call
		path (str): The path on the service
		req (dict): The request details, which can include 'data' and 'session'

	Returns:
		Response
	"""
	return request(service, 'update', path, req)


# coding=utf8
""" Service

Holds the class used to create services that can be started as rest apps
"""

__author__		= "Chris Nasr"
__copyright__	= "Ouroboros Coding Inc."
__email__		= "chris@ouroboroscoding.com"
__created__		= "2023-03-15"

# Ouroboros imports
import config
from tools import evaluate

# Python imports
import abc
from collections.abc import Callable
import re
from typing import List

# Module imports
from body.external import register_service
from body.errors import DATA_FIELDS
from body.response import ResponseException
from body.rest import REST

class Service(abc.ABC):
	"""Service

	The object to build all services from
	"""

	__noun_regex = re.compile(
		r'([a-z]+(?:_[a-z]+)*)_(create|delete|read|update)'
	)
	"""Regular Expression to match to valid service noun method"""

	additional_routes: List[tuple] | None = None
	"""Additional Routes

	Additional URIs can be passed to the server to handle things outside of
	the Service architecture.

	class SomeService(Service):
		pass

	def custom():
		return "<html />"

	SomeService.additional_routes = [ ( '/custom', 'GET', custom ) ]
	"""

	def __init__(self, name: str | None = None):
		"""Constructor

		Initialises the service

		Arguments:
			name (str): Optional, uses __class__.__name__.lower() if not set

		Returns:
			Service
		"""

		# Store the name
		self._name = name is None \
			and self.__class__.__name__.lower() \
			or name

		# Init the list of available methods
		self._requests = []

		# Go through all the functions found on the service
		for sFunc in dir(self):

			# Check the format of the method name
			oMatch = self.__noun_regex.match(sFunc)

			# If it's a match
			if oMatch:

				# Add it to the list
				self._requests.append({
					'name': oMatch.group(1),
					'action': oMatch.group(2),
					'func': getattr(self, sFunc)
				})

		# Register the service with body
		self._info = register_service(self._name, self)

		# Call reset
		self.reset()

	@staticmethod
	def check_data(data: dict, fields: list):
		"""Check Data

		Checks if `fields` are set in the `data` dictionary. Raises a \
		DATA_FIELDS ResponseException if any of the `fields` or sub-fields are \
		missing

		\# Check '_id' and 'name' exist in req.data

		check_data(req.data, [ '_id', 'name' ])

		\# Check '_id' and 'record.name' exist in req.data

		check_data(req.data, [ '_id', { 'record': [ 'name' ]} ])

		\# Check 'record.name', and 'options.raw' exist in req.data

		check_data(req.data, {
			'record': [ 'name' ],
			'options': [ 'raw' ]
		})

		Arguments:
			data (dict): The dict to check for missing fields
			fields (list | dict): The list of fields to check for

		Raises:
			ResponseException
		"""

		# Check the data
		try:
			evaluate(data, fields)
		except ValueError as e:
			raise ResponseException(error = (
				DATA_FIELDS,
				[ [ f, 'missing' ] for f in e.args ]
			))

	@property
	def name(self) -> str:
		"""Name

		Returns the name of the service

		Returns:
			str
		"""
		return self._name

	@abc.abstractmethod
	def reset(self):
		"""Reset

		Called when the system has been reset, usually by loading new data that
		the instance will need to process/reprocess

		Returns:
			None
		"""
		raise NotImplementedError('Must implement the "reset" method')

	@property
	def requests(self) -> List[dict]:
		"""Requests

		Returns a list of all the requests available in the service

		Returns:
			dict[]
		"""
		return self._requests

	def rest(self,
		additional: List[list] = None,
		on_errors: Callable | None = None
	):
		"""Rest

		Creates a REST instance using the service as the only internally
		accessible one

		Additional URIs can be passed to the server to handle things outside of
		the Service architecture.

		def custom():
			return "<html />"

		rest(
			additional = [ [ '/custom', 'GET', custom ] ]
		)

		Arguments:
			additional (list): A list of tuples representing arguments to route.
			on_errors (callable): An optional callback for when a service
				request crashes
		"""

		# Create the REST server using the Client instance
		oRest = REST(
			instances = [ self ],
			cors = config.body.rest.allowed('localhost'),
			on_errors = on_errors,
			verbose = config.body.rest.verbose(False)
		)

		# If there's any additional
		if additional:
			for l in additional:
				oRest.route(*l)

		# Run the server forever
		oRest.run(
			host = self._info['host'],
			port = self._info['port'],
			workers = self._info['workers'],
			timeout = 'timeout' in self._info and self._info['timeout'] or 30
		)


# coding=utf8
""" Errors

Shared error codes
"""

__author__		= "Chris Nasr"
__copyright__	= "Ouroboros Coding Inc."
__email__		= "chris@ouroboroscoding.com"
__created__		= "2022-08-29"

REST_REQUEST_DATA = 100
REST_CONTENT_TYPE = 101
REST_AUTHORIZATION = 102
REST_LIST_TO_LONG = 103
REST_LIST_INVALID_URI = 104
"""REST related errors"""

SERVICE_ACTION = 200
SERVICE_STATUS = 201
SERVICE_CONTENT_TYPE = 202
SERVICE_UNREACHABLE = 203
SERVICE_NOT_REGISTERED = 204
SERVICE_NO_SUCH_NOUN = 205
SERVICE_TO_BE_USED_LATER = 206
SERVICE_CRASHED = 207
SERVICE_NO_DATA = 208
SERVICE_NO_SESSION = 209
"""Service related errors"""

RIGHTS = 1000
"""Rights insufficient or missing"""

DATA_FIELDS = 1001
"""One or more data fields is missing or invalid"""

ALREADY_DONE = 1002
"""Indicates the request has already been done, and can't be done again"""

DB_NO_RECORD = 1100
DB_DUPLICATE = 1101
DB_CREATE_FAILED = 1102
DB_DELETE_FAILED = 1103
DB_UPDATE_FAILED = 1104
DB_KEY_BEING_USED = 1105
DB_ARCHIVED = 1106
DB_REFERENCES = 1107
"""DB related errors"""

__all__ = [ n for n,v in globals().items() if isinstance(v, int) ]
""" Export all the constants"""


# coding=utf8
""" Constants

Shared constants
"""

__author__		= "Chris Nasr"
__copyright__	= "Ouroboros Coding Inc."
__email__		= "chris@ouroboroscoding.com"
__created__		= "2023-03-09"

EMPTY_TUUID = '00000000000040008000000000000000'
"""Empty, but valid, Trimmed UUID"""
EMPTY_UUID = '00000000-0000-4000-8000-000000000000'
"""Empty, but valid, Full UUID"""

SECONDS_HOUR = 3600
SECONDS_DAY = 86400
SECONDS_WEEK = 604800
"""Seconds related constants"""